/*
 * Copyright (C) 2005 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License.
 * You must obey the GNU General Public License in all respects for all of the code used
 * other than the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package mir.entity.adapter;

import mir.entity.Entity;
import mir.util.CachingRewindableIterator;
import mir.util.RewindableIterator;

import java.util.*;

/** An EntityAdapter is a wrapper around an Entity meant to add
 * missing functionality. It provides "calculated fields" 
 * which mir installations can extend reasonably easilly.
 * "calculated fields" compute values that are not directly present
 * in the db table. For example:  
 * a field to have the number of comments associated with an article;
 * a field to get the list of hidden comments associated with an article;
 * etc.
 * <p>This whole framework is meant to be replaced by hibernate</p>
 */
public class EntityAdapter {
  private Entity entity;
  private EntityAdapterDefinition definition;
  private Map calculatedFieldsCache;
  private EntityAdapterModel model;

  public EntityAdapter(Entity anEntity, EntityAdapterDefinition aDefinition, EntityAdapterModel aModel) {
    entity = anEntity;
    definition = aDefinition;
    calculatedFieldsCache = new HashMap();
    model = aModel;
  }

  public boolean equals(Object anObject) {
    return        anObject instanceof EntityAdapter
           && ((EntityAdapter) anObject).entity.equals(entity);
  }

  public int hashCode() {
    if (entity!=null) {
      return entity.hashCode();
    }
    else {
      return 0;
    }
  }

  public Entity getEntity() {
    return entity;
  }

  public EntityAdapterModel getModel() {
    return model;
  }

  public Object get(String aKey) {
    Object result;

    if (calculatedFieldsCache.containsKey(aKey)) {
      return calculatedFieldsCache.get(aKey);
    }
    else if (definition.hasCalculatedField(aKey)) {
      result = definition.getCalculatedField(aKey).getValue(this);
      calculatedFieldsCache.put(aKey, result);

      return result;
    }
    else {
      return entity.getFieldValue(aKey);
    }
  }

  public Iterator getIterator(String aKey) {
    Object result = get(aKey);

    if (result instanceof RewindableIterator) {
      ((RewindableIterator) result).rewind();
      return (RewindableIterator) result;
    }
    else if (result instanceof Iterator) {
      return (Iterator) result;
    }
    else if (result instanceof Collection) {
      return ((Collection) result).iterator();
    }
    else if (result!=null) {
      return Collections.singletonList(result).iterator();
    }
    else {
      return null;
    }
  }

  public Object getComplexRelation(String aMainTablePrefix, List someExtraTables,
    String aWhereClause, String anOrderByClause, String aDefinition) {
    try {
      return
          new CachingRewindableIterator(
            new EntityIteratorAdapter(aMainTablePrefix, someExtraTables,
                aWhereClause, anOrderByClause,
                -1, getModel(), aDefinition, -1, 0)
            );
    }
    catch (Throwable t) {
      throw new RuntimeException(t.getMessage());
    }
  }


  public List getRelation(String aWhereClause, String anOrderByClause, String aDefinition) {
    try {
      return EntityAdapterEngine.retrieveAdapterList(model, aDefinition, aWhereClause, anOrderByClause, -1, 0);
    }
    catch (Throwable t) {
      throw new RuntimeException(t.getMessage());
    }
  }

  public EntityAdapter getToOneRelation(String aWhereClause, String anOrderByClause, String aDefinition) {
    try {
      Iterator i = new EntityIteratorAdapter(aWhereClause, anOrderByClause, 1, getModel(), aDefinition, 1, 0);

      if (i.hasNext())
        return (EntityAdapter) i.next();
			return null;
    }
    catch (Throwable t) {
      throw new RuntimeException(t.getMessage());
    }
  }

  public EntityAdapter getComplexToOneRelation(String aMainTablePrefix, List someExtraTables,
                                               String aWhereClause, String anOrderByClause, String aDefinition) {
    try {
      Iterator i = new EntityIteratorAdapter(aMainTablePrefix, someExtraTables, aWhereClause, anOrderByClause, -1,
          getModel(), aDefinition, 1, 0);

      if (i.hasNext())
        return (EntityAdapter) i.next();
			return null;
    }
    catch (Throwable t) {
      throw new RuntimeException(t.getMessage());
    }
  }

}