/*
 * Copyright (C) 2001, 2002 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License,
 * The Sun (tm) Java Advanced Imaging library (JAI), The Sun JIMI library
 * (or with modified versions of the above that use the same license as the above),
 * and distribute linked combinations including the two.  You must obey the
 * GNU General Public License in all respects for all of the code used other than
 * the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package mir.producer;

import mir.entity.adapter.EntityAdapterModel;
import mir.entity.adapter.EntityIteratorAdapter;
import mir.util.ParameterExpander;
import mir.util.StringRoutines;

import java.util.*;

/**
 * This producer makes it possible to show articles in batches, like on archive pages.
 * <emph> The order by clause should lead to a result set in <b>reverse order</b>:
 *         the first row will be the last entity in the last batch </emph>
 */

public class EntityBatchingProducerNode implements ProducerNode {
  private String batchInfoKey;
  private String batchDataKey;
  private EntityAdapterModel model;
  private String mainTablePrefix;
  private String extraTables;
  private String definition;
  private String whereClause;
  private String orderByClause;
  private String nrEntitiesToSkipExpression;
  private String nrEntitiesPerBatchExpression;
  private String minNrEntitiesInFirstBatchExpression;
  private String nrBatchesToProcessExpression;
  private ProducerNode batchSubNode;
  private ProducerNode batchListSubNode;

  public EntityBatchingProducerNode(
        String aBatchDataKey,
        String aBatchInfoKey,
        EntityAdapterModel aModel,
        String aMainTablePrefix,
        String someExtraTables,
        String aDefinition,
        String aWhereClause,
        String anOrderByClause,
        String anrEntitiesPerBatchExpression,
        String aminNrEntitiesInFirstBatchExpression,
        String anrEntitiesToSkipExpression,
        String aNrBatchesToProcessExpression,
        ProducerNode aBatchSubNode,
        ProducerNode aBatchListSubNode) {

    batchSubNode = aBatchSubNode;
    batchListSubNode = aBatchListSubNode;

    batchDataKey = aBatchDataKey;
    batchInfoKey = aBatchInfoKey;
    model = aModel;
    mainTablePrefix = aMainTablePrefix;
    extraTables = someExtraTables;
    definition = aDefinition;
    whereClause = aWhereClause;
    orderByClause = anOrderByClause;
    nrEntitiesToSkipExpression = anrEntitiesToSkipExpression;
    nrEntitiesPerBatchExpression = anrEntitiesPerBatchExpression;
    minNrEntitiesInFirstBatchExpression = aminNrEntitiesInFirstBatchExpression;
    nrBatchesToProcessExpression = aNrBatchesToProcessExpression;
  }

  protected boolean isAborted(Map aValueMap) {
    Object producerValue = aValueMap.get(NodedProducer.PRODUCER_KEY);
    return (
       (producerValue instanceof NodedProducer) &&
      ((NodedProducer) producerValue).getIsAborted());
  }

  public void produce(ProductionContext aProducerContext) throws ProducerExc, ProducerFailure {
    int nrEntities;
    int nrBatchesAfterFirst;
    int nrEntitiesInFirstBatch;
    int nrBatchesToProcess;
    List batchesData;
    int i;
    Map batchData;
    String expandedWhereClause;
    String expandedOrderByClause;

    List batchLocations;
    BatchLocation location;

    int nrEntitiesToSkip;
    int nrEntitiesPerBatch;
    int minNrEntitiesInFirstBatch;

    try {
      nrBatchesToProcess = ParameterExpander.evaluateIntegerExpressionWithDefault(aProducerContext.getValueSet(), nrBatchesToProcessExpression, -1);

      expandedWhereClause = ParameterExpander.expandExpression( aProducerContext.getValueSet(), whereClause );
      expandedOrderByClause = ParameterExpander.expandExpression( aProducerContext.getValueSet(), orderByClause );

      nrEntitiesToSkip = ParameterExpander.evaluateIntegerExpression( aProducerContext.getValueSet(), nrEntitiesToSkipExpression);
      nrEntitiesPerBatch = ParameterExpander.evaluateIntegerExpression( aProducerContext.getValueSet(), nrEntitiesPerBatchExpression);
      minNrEntitiesInFirstBatch = ParameterExpander.evaluateIntegerExpression( aProducerContext.getValueSet(), minNrEntitiesInFirstBatchExpression);
      List extraTableList = StringRoutines.splitString(ParameterExpander.expandExpression( aProducerContext.getValueSet(), extraTables).trim(), ",");

      batchesData = new ArrayList();
      batchLocations = new ArrayList();

      nrEntities = model.getMappingForName(definition).getDatabase().getSize(
          mainTablePrefix, extraTableList, expandedWhereClause)-nrEntitiesToSkip;
      nrEntitiesInFirstBatch = nrEntities % nrEntitiesPerBatch;
      while (nrEntitiesInFirstBatch<minNrEntitiesInFirstBatch && nrEntities-nrEntitiesInFirstBatch>=nrEntitiesPerBatch)
        nrEntitiesInFirstBatch = nrEntitiesInFirstBatch + nrEntitiesPerBatch;
      nrBatchesAfterFirst = (nrEntities-nrEntitiesInFirstBatch)/nrEntitiesPerBatch;

      batchLocations.add(new BatchLocation(nrBatchesAfterFirst*nrEntitiesPerBatch, nrEntitiesInFirstBatch));
      batchData = new HashMap();
      batchData.put("identifier", "");
      batchData.put("index", new Integer(nrBatchesAfterFirst+1));
      batchData.put("size", new Integer(nrEntitiesInFirstBatch));
      batchesData.add(batchData);

      for (i=0; i<nrBatchesAfterFirst; i++) {
        batchLocations.add(1, new BatchLocation(i*nrEntitiesPerBatch, nrEntitiesPerBatch));
        batchData = new HashMap();
        batchData.put("identifier", new Integer(i));
        batchData.put("index", new Integer(i+1));
        batchData.put("size", new Integer(nrEntitiesPerBatch));
        batchesData.add(1, batchData);
      }

      batchData = new HashMap();
      ParameterExpander.setValueForKey(aProducerContext.getValueSet(), batchInfoKey, batchData);
      batchData.put("all", batchesData);
      batchData.put("first", batchesData.get(0));
      batchData.put("last", batchesData.get(batchesData.size()-1));
      batchData.put("count", new Integer(batchesData.size()));

      if (batchListSubNode!=null && (!isAborted(aProducerContext.getValueSet()))) {
        batchListSubNode.produce(aProducerContext);
      }

      if (nrBatchesToProcess<0 || nrBatchesToProcess>nrBatchesAfterFirst+1) {
        nrBatchesToProcess = nrBatchesAfterFirst+1;
      }

      if (batchSubNode!=null) {
        for (i=0; i<nrBatchesToProcess && !isAborted(aProducerContext.getValueSet()); i++) {
          location = (BatchLocation) batchLocations.get(i);

          batchData.put("current", batchesData.get(i));
          if (i>0)
            batchData.put("previous", batchesData.get(i-1));
          else
            batchData.put("previous", null);

          if (i<batchesData.size()-1)
            batchData.put("next", batchesData.get(i+1));
          else
            batchData.put("next", null);

          Iterator j = new EntityIteratorAdapter(mainTablePrefix, extraTableList, expandedWhereClause, expandedOrderByClause,
                    location.nrEntities, model, definition, location.nrEntities, location.firstEntity);
          List entities = new ArrayList();

          while (j.hasNext())
            entities.add(0, j.next());

          ParameterExpander.setValueForKey(aProducerContext.getValueSet(), batchDataKey, entities );

          batchSubNode.produce(aProducerContext);
        }
      }
    }
    catch (Throwable t) {
      aProducerContext.getLogger().warn("EntityBatchingProducerNode caused an exception: " + t.getMessage());
    }
  }

  private class BatchLocation {
    int nrEntities;
    int firstEntity;

    public BatchLocation(int aFirstEntity, int aNrEntities) {
      firstEntity = aFirstEntity;
      nrEntities = aNrEntities;
    }
  }
}

