/*
 * Copyright (C) 2001, 2002 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License,
 * The Sun (tm) Java Advanced Imaging library (JAI), The Sun JIMI library
 * (or with modified versions of the above that use the same license as the above),
 * and distribute linked combinations including the two.  You must obey the
 * GNU General Public License in all respects for all of the code used other than
 * the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package mir.producer;

import mir.log.LoggerWrapper;
import mir.util.ParameterExpander;

import java.util.Map;

public class ExecuteProgramProducerNode extends AbstractProducerNode  {
  private String scriptExpression;
  private String maxDurationExpression;
  private String outputVariableExpression;
  private String returnValueVariableExpression;

  public ExecuteProgramProducerNode(String aScriptExpression, String aMaxDurationExpression, String anOutputVariableExpression,
                                    String aReturnValueVariableExpression) {
    scriptExpression = aScriptExpression;
    maxDurationExpression = aMaxDurationExpression;
    outputVariableExpression = anOutputVariableExpression;
    returnValueVariableExpression = aReturnValueVariableExpression;
  }

  public void produce(Map aValueMap, String aVerb, LoggerWrapper aLogger) throws ProducerFailure {
    String script;
    long maxDuration;

    try {
      script = ParameterExpander.expandExpression(aValueMap, scriptExpression);
      maxDuration = ParameterExpander.evaluateIntegerExpressionWithDefault(aValueMap, maxDurationExpression, 0);

      ProcessRunner runner = new ProcessRunner(aLogger, script);
      runner.start();

      synchronized (runner) {
        runner.wait(maxDuration);
      }
      runner.interrupt();

      if (runner.getFinished()) {
        aLogger.info(script + " terminated successfully, return value = " + runner.getReturnValue() + ".");

        if (returnValueVariableExpression != null) {
          ParameterExpander.setValueForKey(aValueMap,
              ParameterExpander.expandExpression(aValueMap, returnValueVariableExpression),
              new Integer(runner.getReturnValue()));
        }

      }
      else {
        aLogger.info(script + " interrupted prematurely after " + maxDuration + "ms.");
      }
    }
    catch (Throwable e) {
      aLogger.error("Error while executing " + scriptExpression + " : " + e.toString());
    }
  }

  private static class ProcessRunner extends Thread {
    private String script;
    private boolean finished = false;
    private int returnValue = 0;
    private LoggerWrapper logger;

    public ProcessRunner(LoggerWrapper aLogger, String aScript) {
      script = aScript;
      logger = aLogger;
    }

    public boolean getFinished() {
      return finished;
    }

    public int getReturnValue() {
      return returnValue;
    }

    public void run() {
      Process process = null;
      try {
        process = Runtime.getRuntime().exec(script);
        returnValue = process.waitFor();

        finished = true;

        synchronized (this) {
          this.notify();
        }
      }
      catch (InterruptedException e) {
        if (process!=null) {
          process.destroy();
        }
      }
      catch (Exception e) {
        logger.error(script + " failed to execute: " + e.getMessage());
      }
    }
  }

}
