/*
 * Copyright (C) 2001, 2002 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License,
 * The Sun (tm) Java Advanced Imaging library (JAI), The Sun JIMI library
 * (or with modified versions of the above that use the same license as the above),
 * and distribute linked combinations including the two.  You must obey the
 * GNU General Public License in all respects for all of the code used other than
 * the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package mir.servlet;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.sql.Connection;
import java.util.Locale;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import mir.config.MirPropertiesConfiguration;
import mir.log.LoggerWrapper;
import mircoders.global.MirGlobal;

/**
 * Title:        Mir
 * Description:  Abstract servlet-class
 * Copyright:    Copyright (c) 2001, 2002
 * Company:      Mir-coders group
 * @author       idfx, the Mir-coders group
 * @version      $Id: AbstractServlet.java,v 1.30.2.10 2005/10/01 18:40:40 zapata Exp $
 */

public abstract class AbstractServlet extends HttpServlet {
  protected LoggerWrapper logger;
  protected MirPropertiesConfiguration configuration;

  public AbstractServlet() {
    super();
  }

  protected void setNoCaching(HttpServletResponse aResponse) {
    aResponse.setHeader("Pragma", "no-cache");
    aResponse.setDateHeader("Expires", 0);
    aResponse.setHeader("Cache-Control", "no-cache");
  }

  /**
   * Bind the language to the session
   */
  protected void setLanguage(HttpSession aRequest, String aLanguage) {
    aRequest.setAttribute("language", aLanguage);
    aRequest.setAttribute("locale", new Locale(aLanguage, ""));
  }

  /**
   * get the locale either from the session or the accept-language header ot the request
   * this supersedes getLanguage for the new i18n
   */
  public Locale getLocale(HttpServletRequest aRequest) {
    Locale locale = null;
    HttpSession session = aRequest.getSession(false);
    if (session != null) {
      // session can be null in case of logout
      locale = (Locale) session.getAttribute("locale");
    }
    // if there is nothing in the session get it fron the accept-language
    if (locale == null) {
      locale = aRequest.getLocale();
    }

    logger.debug("getting locale: " + locale.getLanguage());

    return locale;
  }

  /**
   * Checks the Accept-Language of the client browser.
   * If this language is available it returns its country-code,
   * else it returns the standard-language
   */
  protected String getAcceptLanguage(HttpServletRequest aRequest) {
    Locale locale = aRequest.getLocale();

    return locale.getLanguage();
  }

  /**
   * @see javax.servlet.Servlet#init(javax.servlet.ServletConfig)
   */
  public void init(ServletConfig aConfiguration) throws ServletException {
    super.init(aConfiguration);

    MirPropertiesConfiguration.setContext(aConfiguration.getServletContext());
    try {
      configuration = MirPropertiesConfiguration.instance();
    }
    catch (Throwable t) {
      throw new ServletException("can't read configuration: " + t.toString());
    }

    logger = new LoggerWrapper("Servlet");

    try {
      MirGlobal.localizer();
    }
    catch (Throwable t) {
      logger.fatal("can't get localizer: " + t.toString());
      throw new ServletException("can't get localizer: " + t.toString());
    }

    try {
      Connection connection = MirGlobal.getDatabaseEngine().obtainConnection();
      MirGlobal.getDatabaseEngine().releaseConnection(connection);
    }
    catch (Throwable t) {
      logger.fatal("Can't connect to database: " + t.toString());
      throw new ServletException("Can't connect to database: " + t.toString());
    }
  }

  private void setEncoding(HttpServletRequest aRequest) {
    try {
      logger.info("Request has encoding: " + aRequest.getCharacterEncoding());
      logger.info("Config stipulates encoding: " + configuration.getString("Mir.DefaultHTMLCharset"));

      Class requestClass = aRequest.getClass();
      Method method = requestClass.getMethod("setCharacterEncoding", new Class[]{String.class});
      String encoding = configuration.getString("Mir.DefaultHTMLCharset");
      method.invoke(aRequest, new Object[]{encoding});
      logger.info("Request now has encoding: " + aRequest.getCharacterEncoding());
    }
    catch (NoSuchMethodException e) {
    }
    catch (SecurityException e) {
      logger.error(e.getMessage(), e);
    }
    catch (IllegalArgumentException e) {
      logger.error(e.getMessage(), e);
    }
    catch (IllegalAccessException e) {
      logger.error(e.getMessage(), e);
    }
    catch (InvocationTargetException e) {
      logger.error(e.getMessage(), e);
    }
  }

  protected final void doGet(HttpServletRequest aRequest, HttpServletResponse aResponse) throws ServletException, IOException {
    doPost(aRequest, aResponse);
  }

  protected final void doPost(HttpServletRequest aRequest, HttpServletResponse aResponse) throws ServletException, IOException {
    if ((configuration.getString("RootUri") == null) ||
         configuration.getString("RootUri").equals("")) {
      configuration.setProperty("RootUri", aRequest.getContextPath());
    }

    setEncoding(aRequest);
    process(aRequest, aResponse);
  }

  public abstract void process(HttpServletRequest aRequest, HttpServletResponse aResponse) throws ServletException, IOException;

  /**
   * Selects the language for the response.
   */
  protected void checkLanguage(HttpSession aSession, HttpServletRequest aRequest) {
    String requestLanguage = aRequest.getParameter("language");
    String sessionLanguage = (String) aSession.getAttribute("language");
    String acceptLanguage = aRequest.getLocale().getLanguage();
    String defaultLanguage = configuration.getString("Mir.Login.DefaultLanguage", "en");

    String language = requestLanguage;

    if (language==null)
      language = sessionLanguage;

    if (language==null)
      language = acceptLanguage;

    if (language==null)
      language = defaultLanguage;

    setLanguage(aSession, language);
  }
}
