/*
 * Copyright (C) 2001, 2002 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License,
 * The Sun (tm) Java Advanced Imaging library (JAI), The Sun JIMI library
 * (or with modified versions of the above that use the same license as the above),
 * and distribute linked combinations including the two.  You must obey the
 * GNU General Public License in all respects for all of the code used other than
 * the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package mir.servlet;

import mir.entity.adapter.EntityAdapterEngine;
import mir.entity.adapter.EntityAdapterModel;
import mir.module.AbstractModule;
import mir.module.ModuleExc;
import mir.storage.Database;
import mir.util.HTTPRequestParser;
import mir.util.URLBuilder;
import mir.util.StringRoutines;
import mircoders.global.MirGlobal;
import mircoders.localizer.MirLocalizerExc;
import mircoders.servlet.ServletHelper;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.io.IOException;

/**
 *
 */

public abstract class AdminServletModule extends ServletModule {
  protected AbstractModule mainModule;
  protected String definition;
  private EntityAdapterModel model;

  protected String listGenerator;
  protected String editGenerator;
  protected String deleteConfirmationGenerator;
  protected int nrEntitiesPerListPage;

  /** the list of parameters that need to be propagated in the list */
  private Set propagatedParameters = new HashSet();

  protected void addPropagatedParameter(String aParameter) {
    propagatedParameters.add(aParameter);
  }

  protected AdminServletModule(){
    addPropagatedParameter("searchfield");
    addPropagatedParameter("searchtext");
    addPropagatedParameter("searchispublished");
    addPropagatedParameter("searchstatus");
    addPropagatedParameter("searchorder");

    definition = null;
    try {
      model = MirGlobal.localizer().dataModel().adapterModel();
    }
    catch (MirLocalizerExc e) {
      getLogger().error("Can't create model: " + e.toString());

      throw new ServletModuleFailure("Can't retrieve model", e);
    }

    listGenerator = getConfiguration().getString("ServletModule." + getName() + ".ListTemplate");
    editGenerator = getConfiguration().getString("ServletModule." + getName() + ".EditTemplate");
    deleteConfirmationGenerator = getConfiguration().getString("ServletModule." + getName() + ".DeleteConfirmationTemplate");
    nrEntitiesPerListPage =
        getConfiguration().getInt("ServletModule."+getName()+".ListSize",
        getConfiguration().getInt("ServletModule.Default.ListSize", 20));

  }

  protected EntityAdapterModel getModel() {
    return model;
  }

  protected void logAdminUsage(HttpServletRequest aRequest, String anObject, String aDescription) {
    MirGlobal.logAdminUsage(ServletHelper.getUser(aRequest), getName() + ":" + anObject, aDescription);
  }

  /**
   * Function to specify the default ordering for lists. May be overridden.
   */
  protected String getDefaultListOrdering() {

    if (mainModule!=null && mainModule.getStorageObject()!=null &&
        mainModule.getStorageObject().hasField("webdb_create")) {
      return "webdb_create desc";
    }
    else {
      return "id asc";
    }
  }

  /**
   * Generic list servlet method
   */

  public void list(HttpServletRequest aRequest, HttpServletResponse aResponse) throws ServletModuleExc {
    HTTPRequestParser requestParser = new HTTPRequestParser(aRequest);

    String where = requestParser.getParameter("where");
    String order = requestParser.getParameterWithDefault("order", getDefaultListOrdering());
    int offset = requestParser.getIntegerWithDefault("offset", 0);
    String mainTablePrefix = requestParser.getParameter("mainTablePrefix");
    String extraTables = requestParser.getParameter("extraTables");

    returnList(aRequest, aResponse, where, order, offset, Collections.EMPTY_MAP, mainTablePrefix, extraTables);
  }


  public void returnList(HttpServletRequest aRequest, HttpServletResponse aResponse,
                         String aWhereClause, String anOrderByClause, int anOffset) throws ServletModuleExc {
    returnList(aRequest, aResponse, aWhereClause, anOrderByClause, anOffset, Collections.EMPTY_MAP);
  }

  public void returnList(HttpServletRequest aRequest, HttpServletResponse aResponse,
                         String aWhereClause, String anOrderByClause, int anOffset,
                         Map anOverridingRequestParameters) throws ServletModuleExc {
    returnList(aRequest, aResponse, aWhereClause, anOrderByClause, anOffset, anOverridingRequestParameters, "", null);
  }
  public void returnList(HttpServletRequest aRequest, HttpServletResponse aResponse,
                         String aWhereClause, String anOrderByClause, int anOffset,
                         Map anOverridingRequestParameters, String aMainTablePrefix, String someExtraTables) throws ServletModuleExc {
    HTTPRequestParser requestParser = new HTTPRequestParser(aRequest, anOverridingRequestParameters);
    URLBuilder urlBuilder = new URLBuilder();

    try {
      Map responseData = ServletHelper.makeGenerationData(aRequest, aResponse, getLocales(aRequest));

      List extraTables = null;
      if (someExtraTables!=null && someExtraTables.length() > 0) {
        extraTables = StringRoutines.splitString(someExtraTables, ",");
      }

      List list =
         EntityAdapterEngine.retrieveAdapterList(model, definition, aMainTablePrefix, extraTables, aWhereClause, anOrderByClause, nrEntitiesPerListPage, anOffset);

      responseData.put("nexturl", null);
      responseData.put("prevurl", null);
      responseData.put("module", getName());

      urlBuilder.setValue("module", getName());
      urlBuilder.setValue("do", "list");
      urlBuilder.setValue("where", aWhereClause);
      urlBuilder.setValue("order", anOrderByClause);
      urlBuilder.setValue("extraTables", someExtraTables);
      urlBuilder.setValue("mainTablePrefix", aMainTablePrefix);


      urlBuilder.setValue("offset", anOffset);
      responseData.put("offset" , Integer.toString(anOffset));
      responseData.put("thisurl" , urlBuilder.getQuery());

      propagateFields(requestParser, urlBuilder, responseData);

      if (list.size()>=nrEntitiesPerListPage) {
        urlBuilder.setValue("offset", anOffset + nrEntitiesPerListPage);
        responseData.put("nexturl" , urlBuilder.getQuery());
      }

      if (anOffset>0) {
        urlBuilder.setValue("offset", Math.max(anOffset - nrEntitiesPerListPage, 0));
        responseData.put("prevurl" , urlBuilder.getQuery());
      }

      responseData.put("entities", list);
      responseData.put("from" , Integer.toString(anOffset+1));
      responseData.put("count", "?");
      responseData.put("to", Integer.toString(anOffset+list.size()-1));

      ServletHelper.generateResponse(aResponse.getWriter(), responseData, listGenerator);
    }
    catch (Throwable e) {
      throw new ServletModuleFailure(e);
    }
  }

  public void editObject(HttpServletRequest aRequest, HttpServletResponse aResponse, String anId) throws ServletModuleExc {
    try {
      editObject(aRequest, aResponse, model.makeEntityAdapter(definition, mainModule.getById(anId)), false, anId);
    }
    catch (Throwable t) {
      throw new ServletModuleFailure(t);
    }
  }

  public void editObject(HttpServletRequest aRequest, HttpServletResponse aResponse, Object anObject, boolean anIsNew, String anId) throws ServletModuleExc {
    HTTPRequestParser requestParser = new HTTPRequestParser(aRequest);
    URLBuilder urlBuilder = new URLBuilder();

    try {
      Map responseData = ServletHelper.makeGenerationData(aRequest, aResponse, getLocales(aRequest));

      responseData.put("module", getName());
      responseData.put("entity", anObject);
      responseData.put("new", new Boolean(anIsNew));


      urlBuilder.setValue("module", getName());
      urlBuilder.setValue("returnurl", requestParser.getParameter("returnurl"));
      if (anIsNew) {
        urlBuilder.setValue("do", "add");
      }
      else {
        urlBuilder.setValue("id", anId);
        urlBuilder.setValue("do", "edit");
      }
      responseData.put("returnurl", requestParser.getParameter("returnurl"));
      responseData.put("thisurl", urlBuilder.getQuery());

      ServletHelper.generateResponse(aResponse.getWriter(), responseData, editGenerator);
    }
    catch (Throwable e) {
      throw new ServletModuleFailure(e);
    }
  }

  /**
   * Generic add servlet method
   */
  public void add(HttpServletRequest aRequest, HttpServletResponse aResponse)
      throws ServletModuleExc, ServletModuleUserExc, ServletModuleFailure  {

    Map object = new HashMap();

    Iterator i = mainModule.getStorageObject().getFieldNames().iterator();

    while (i.hasNext()) {
      object.put(i.next(), "");
    }

    initializeNewObject(object, aRequest, aResponse);

    editObject(aRequest, aResponse, object, true, null);
  }

  protected void initializeNewObject(Map aNewObject, HttpServletRequest aRequest, HttpServletResponse aResponse) {
  }

  /**
   * Generic edit servlet method
   */
  public void edit(HttpServletRequest aRequest, HttpServletResponse aResponse) throws ServletModuleExc, ServletModuleUserExc, ServletModuleFailure  {
    edit(aRequest, aResponse, aRequest.getParameter("id"));
  }

  /**
   */
  public void edit(HttpServletRequest aRequest, HttpServletResponse aResponse, String anIdentifier)
      throws ServletModuleExc, ServletModuleUserExc, ServletModuleFailure  {
    try {
      editObject(aRequest, aResponse, model.makeEntityAdapter(definition, mainModule.getById(anIdentifier)), false, anIdentifier);
    }
    catch (Throwable e) {
      throw new ServletModuleFailure(e);
    }
  }

  /**
   * Generic update servlet method
   */
  public void update(HttpServletRequest aRequest, HttpServletResponse aResponse)
      throws ServletModuleExc, ServletModuleUserExc, ServletModuleFailure  {
    try {
      HTTPRequestParser requestParser = new HTTPRequestParser(aRequest);

      String id = aRequest.getParameter("id");
      Map withValues = getIntersectingValues(aRequest, mainModule.getStorageObject());
      mainModule.set(withValues);

      logAdminUsage(aRequest, id, "object modified");

      String returnUrl = requestParser.getParameter("returnurl");

      if (returnUrl!=null) {
        ServletHelper.redirect(aResponse, returnUrl);
      }
      else {
        edit(aRequest, aResponse, id);
      }
    }
    catch (Throwable e) {
      throw new ServletModuleFailure(e);
    }
  }

  /**
   * Generic insert servlet method
   */
  public void insert(HttpServletRequest aRequest, HttpServletResponse aResponse)
      throws ServletModuleExc, ServletModuleUserExc, ServletModuleFailure  {
    try {
      HTTPRequestParser requestParser = new HTTPRequestParser(aRequest);

      Map object = getIntersectingValues(aRequest, mainModule.getStorageObject());

      String id = processInstertedObject(object, aRequest, aResponse);

      logAdminUsage(aRequest, id, "object inserted");

      String returnUrl = requestParser.getParameter("returnurl");

      if (returnUrl!=null) {
        ServletHelper.redirect(aResponse, returnUrl);
      }
      else {
        edit(aRequest, aResponse, id);
      }
    }
    catch (Throwable e) {
      throw new ServletModuleFailure(e);
    }
  }

  /**
   *
   */
  public String processInstertedObject(Map anObject, HttpServletRequest aRequest, HttpServletResponse aResponse) {
    try {
      return mainModule.add(anObject);
    }
    catch (ModuleExc t) {
      throw new ServletModuleFailure(t);
    }
  }

  /**
   * Generic delete confirmation servlet method
   */
  public void confirmdelete(HttpServletRequest aRequest, HttpServletResponse aResponse) {
    try {
      String idParam = aRequest.getParameter("id");
      String confirmParam = aRequest.getParameter("confirm");

      if (confirmParam != null && !"".equals(confirmParam)) {
        mainModule.deleteById(idParam);
        logAdminUsage(aRequest, idParam, "object deleted");
        ServletHelper.redirect(aResponse, aRequest.getParameter("okurl"));
      }
      else {
        ServletHelper.redirect(aResponse, aRequest.getParameter("cancelurl"));
      }
    }
    catch (Throwable t) {
      throw new ServletModuleFailure(t);
    }
  }

  /**
   * Generic delete servlet method
   */
  public void delete(HttpServletRequest aRequest, HttpServletResponse aResponse)
      throws ServletModuleExc, ServletModuleUserExc, ServletModuleFailure  {
    String idParam = aRequest.getParameter("id");

    if (idParam == null) {
      throw new ServletModuleExc("Invalid call to delete: no id supplied");
    }

    Map responseData = ServletHelper.makeGenerationData(aRequest, aResponse, getLocales(aRequest));

    responseData.put("module", getName());
    responseData.put("id", idParam);
    responseData.put("cancelurl", aRequest.getParameter("cancelurl"));
    responseData.put("okurl", aRequest.getParameter("okurl"));

    try {
      ServletHelper.generateResponse(aResponse.getWriter(), responseData, deleteConfirmationGenerator);
    }
    catch (IOException e) {
      throw new ServletModuleFailure(e);
    }
  }

  /**
   * Gets the fields from a httprequest and matches them with the metadata from
   * the database object. Returns the keys that match, with their values.
   *
   * @return Map with the values
   */
  public Map getIntersectingValues(HttpServletRequest aRequest, Database aDatabase)
      throws ServletModuleExc, ServletModuleFailure {

    HTTPRequestParser parser = new HTTPRequestParser(aRequest);

    List fields = aDatabase.getFieldNames();

    Map result = new HashMap();
    for (int i = 0; i < fields.size(); i++) {
      String aField = (String) fields.get(i);

      String aValue = parser.getParameter(aField);
      if (aValue != null) {
        result.put(aField, aValue);
      }
    }

    return result;
  }

  private void propagateFields(HTTPRequestParser aRequest, URLBuilder aUrlBuilder, Map aResponseData) {
    Iterator i = propagatedParameters.iterator();
    while (i.hasNext()) {
      String parameter = (String) i.next();
      String value = aRequest.getParameter(parameter);
      aUrlBuilder.setValue(parameter, value);
      aResponseData.put(parameter, value);
    }
  }
}