/*
 * Copyright (C) 2001, 2002 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License,
 * The Sun (tm) Java Advanced Imaging library (JAI), The Sun JIMI library
 * (or with modified versions of the above that use the same license as the above),
 * and distribute linked combinations including the two.  You must obey the
 * GNU General Public License in all respects for all of the code used other than
 * the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package mircoders.media;

import mir.config.MirPropertiesConfiguration;
import mir.entity.Entity;
import mir.log.LoggerWrapper;
import mir.media.MediaExc;
import mir.media.MediaFailure;
import mir.media.MediaHandler;
import mir.misc.StringUtil;
import mir.session.SessionExc;
import mir.session.UploadedFile;
import mir.util.IORoutines;
import mir.util.FileRoutines;
import mircoders.entity.EntityImages;

import java.io.*;
import java.sql.SQLException;

/**
 * This class handles saving, fetching creating representations
 * for all images. The image content is stored in the DB. The content is
 * written out to a file at the ProducerImages level.
 * Remember that Handlers for specific image types, Gif, Jpeg, etc..
 * should override it.
 * It implements the MirMediaHandler interface.
 * <p>
 * slowly starting to look better, a next step would be to have the
 * representation stuff (WebdbImage) happen here.
 * -mh 01.03.2002
 *
 * @see mir.media.MediaHandler
 * @author mh
 * @version $Id: MediaHandlerImages.java,v 1.23.2.10 2006/01/28 18:33:16 zapata Exp $
 */


public abstract class MediaHandlerImages extends AbstractMediaHandler implements MediaHandler {
  protected static MirPropertiesConfiguration configuration = MirPropertiesConfiguration.instance();
  protected static final String PNG = "PNG";
  protected static final String JPEG = "JPEG";

  protected LoggerWrapper logger;

  abstract String getType();

  public MediaHandlerImages() {
    logger = new LoggerWrapper("Media.Images");
  }

  public InputStream getMedia(Entity ent, Entity mediaTypeEnt) throws MediaExc, MediaFailure {
    try {
      return new ByteArrayInputStream(((EntityImages) ent).getImage());
    }
    catch (Throwable e) {
      logger.error("MediaHandlerImages.getImage: " + e.toString());

      throw new MediaFailure(e);
    }
  }

  public void store(UploadedFile anUploadedFile, Entity aMedia, Entity aMediaType) throws MediaExc, MediaFailure {
    try {
      store(anUploadedFile.getInputStream(), aMedia, aMediaType);
    }
    catch (SessionExc e) {
      throw new MediaFailure(e);
    }
  }

  public void store(InputStream in, Entity ent, Entity mediaTypeEnt) throws MediaExc, MediaFailure {
    try {
      ((EntityImages) ent).setImage(in, getType());
    }
    catch (IOException e) {
      logger.error("MediaHandlerImages.store", e);

      throw new MediaFailure("A problem has occurred processing the media file", e);
    }
    catch (SQLException e) {
      logger.error("MediaHandlerImages.store", e);

      throw new MediaFailure("A problem has occurred processing the media file", e);
    }
  }

  public void produce(Entity ent, Entity mediaTypeEnt) throws MediaExc, MediaFailure {
    String date = ent.getFieldValue("date");
    String datePath = StringUtil.webdbDate2path(date);
    String ext = "."+mediaTypeEnt.getFieldValue("name");
    String filepath = datePath+ent.getId()+ext;
    String iconFilePath = configuration.getString("Producer.StorageRoot") + getBaseIconStoragePath() + filepath;
    String productionFilePath = getBaseStoragePath() + File.separator + filepath;


    if (ent.getFieldValue("icon_data")!= null &&
        ent.getFieldValue("image_data")!= null) {
      // make icon
      try {
        OutputStream out;
        InputStream in = new ByteArrayInputStream(((EntityImages) ent).getIcon());
        try {
          File iconFile = new File(iconFilePath);

          FileRoutines.createParentDirectories(iconFile);

          out = new FileOutputStream(iconFile);
          try {
            IORoutines.copyStream(in, out);
          }
          finally {
            out.close();
          }
        }
        finally {
          in.close();
        }

        in = new ByteArrayInputStream(((EntityImages) ent).getImage());
        try {
          File imageFile = new File(productionFilePath);

          FileRoutines.createParentDirectories(imageFile);

          out = new FileOutputStream(imageFile);
          try {
            IORoutines.copyStream(in, out);
          }
          finally {
            out.close();
          }
        }
        finally {
          in.close();
        }

        ent.setFieldValue("icon_path", getBaseIconStoragePath() + filepath);
        ent.setFieldValue("publish_path", filepath);
        ent.update();
      }
      catch (Throwable e) {
        logger.error("Error producing image", e);

        throw new MediaFailure("Error producing image", e);
      }
      finally {
      }
    }
    else {
      logger.error("Can't produce image " + ent.getId() + ": missing image or icon OID");

      throw new MediaExc("Can't produce image " + ent.getId() + ": missing image or icon OID");
    }
  }

  public InputStream getThumbnail(Entity ent) throws MediaExc, MediaFailure {
    try {
      return new ByteArrayInputStream(((EntityImages) ent).getIcon());
    }
    catch (Throwable e) {
      logger.error("MediaHandlerImages.getIcon: " + e.toString());
      throw new MediaFailure(e);
    }
  }

  public String getBaseStoragePath() {
    return configuration.getString("Producer.Image.Path");
  }

  public String getBaseIconStoragePath() {
    return configuration.getString("Producer.Image.IconPath");
  }

  public String getPublishHost() {
    return StringUtil.removeSlash(configuration.getString("Producer.Image.Host"));
  }

  public String getTinyIconName() {
    return configuration.getString("Producer.Icon.TinyImage");
  }

  public String getBigIconName() {
    return configuration.getString("Producer.Icon.BigImage");
  }

  public String getIconAltName() {
    return "Image";
  }

  public String getDescr(Entity mediaType) {
    return "image/jpeg";
  }

}
